import audioop
import shutil
import subprocess
import threading
import time
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Union, Optional, TYPE_CHECKING

__all__ = ("AudioBuffer", "BaseAudio", "Audio", "AudioVolume", "RawInputAudio")

from interactions.client.const import get_logger
from interactions.api.voice.opus import Encoder
from interactions.client.utils import FastJson

if TYPE_CHECKING:
    from interactions.api.voice.recorder import Recorder


class RawInputAudio:
    decoded: bytes
    """The decoded audio"""
    pcm: bytes
    """The raw PCM audio"""
    sequence: int
    """The audio sequence"""
    audio_timestamp: int
    """The current timestamp for this audio"""
    timestamp_ns: float
    """The time this audio was received, in nanoseconds"""
    timestamp: float
    """The time this audio was received, in seconds"""
    ssrc: int
    """The source of this audio"""
    _recoder: "Recorder"
    """A reference to the audio recorder managing this object"""

    def __init__(self, recorder: "Recorder", data: bytes) -> None:
        self.decoded: bytes = b""
        self._recorder = recorder
        self.timestamp_ns = time.monotonic_ns()
        self.timestamp = self.timestamp_ns / 1e9
        self.pcm = b""

        self.ingest(data)

    def ingest(self, data: bytes) -> bytes | None:
        data = bytearray(data)
        header = data[:12]

        decrypted: bytes = self._recorder.decrypt(header, data[12:])
        self.ssrc = int.from_bytes(header[8:12], byteorder="big")
        self.sequence = int.from_bytes(header[2:4], byteorder="big")
        self.audio_timestamp = int.from_bytes(header[4:8], byteorder="big")

        if not self._recorder.recording_whitelist or self.user_id in self._recorder.recording_whitelist:
            # noinspection PyProtectedMember
            if decrypted[0] == 0xBE and decrypted[1] == 0xDE:
                # rtp header extension, remove it
                header_ext_length = int.from_bytes(decrypted[2:4], byteorder="big")
                decrypted = decrypted[4 + 4 * header_ext_length :]
            self.decoded = self._recorder.get_decoder(self.ssrc).decode(decrypted)
            return self.decoded

    @property
    def user_id(self) -> Optional[int]:
        """The ID of the user who made this audio."""
        while not self._recorder.state.ws.user_ssrc_map.get(self.ssrc):
            time.sleep(0.05)
        return self._recorder.state.ws.user_ssrc_map.get(self.ssrc)["user_id"]


class AudioBuffer:
    def __init__(self) -> None:
        self._buffer = bytearray()
        self._lock = threading.Lock()
        self.initialised = threading.Event()

    def __len__(self) -> int:
        return len(self._buffer)

    def extend(self, data: bytes) -> None:
        """
        Extend the buffer with additional data.

        Args:
            data: The data to add
        """
        with self._lock:
            self._buffer.extend(data)

    def read(self, total_bytes: int, *, pad: bool = True) -> bytearray:
        """
        Read `total_bytes` bytes of audio from the buffer.

        Args:
            total_bytes: Amount of bytes to read.
            pad: Whether to pad incomplete frames with 0's.

        Returns:
            Desired amount of bytes

        Raises:
            ValueError: If `pad` is False and the buffer does not contain enough data.
        """
        with self._lock:
            view = memoryview(self._buffer)
            self._buffer = bytearray(view[total_bytes:])
            data = bytearray(view[:total_bytes])
            if 0 < len(data) < total_bytes:
                if pad:
                    # pad incomplete frames with 0's
                    data.extend(b"\0" * (total_bytes - len(data)))
                else:
                    raise ValueError(
                        f"Buffer does not contain enough data to fulfill request {len(data)} < {total_bytes}"
                    )
            return data

    def read_max(self, total_bytes: int) -> bytearray:
        """
        Read up to `total_bytes` bytes of audio from the buffer.

        Args:
            total_bytes: Maximum amount of bytes to read.

        Returns:
            Desired amount of bytes

        Raises:
            EOFError: If the buffer is empty.
        """
        with self._lock:
            if len(self._buffer) == 0:
                raise EOFError("Buffer is empty")
            view = memoryview(self._buffer)
            self._buffer = bytearray(view[total_bytes:])
            return bytearray(view[:total_bytes])


class BaseAudio(ABC):
    """Base structure of the audio."""

    locked_stream: bool
    """Prevents the audio task from closing automatically when no data is received."""
    needs_encode: bool
    """Does this audio data need encoding with opus?"""
    bitrate: Optional[int]
    """Optionally specify a specific bitrate to encode this audio data with"""
    encoder: Optional[Encoder]
    """The encoder to use for this audio data"""

    def __del__(self) -> None:
        self.cleanup()

    @abstractmethod
    def cleanup(self) -> None:
        """A method to optionally cleanup after this object is no longer required."""
        ...

    @property
    def audio_complete(self) -> bool:
        """A property to tell the player if more audio is expected."""
        return False

    @abstractmethod
    def read(self, frame_size: int) -> bytes:
        """
        Reads frame_size ms of audio from source.

        Returns:
            bytes of audio
        """
        ...


class Audio(BaseAudio):
    """Audio for playing from file or URL."""

    source: str
    """The source ffmpeg should use to play the audio"""
    process: subprocess.Popen
    """The ffmpeg process to use"""
    buffer: AudioBuffer
    """The audio objects buffer to prevent stuttering"""
    buffer_seconds: int
    """How many seconds of audio should be buffered"""
    read_ahead_task: threading.Thread
    """A thread that reads ahead to create the buffer"""
    probe: bool
    """Should ffprobe be used to detect audio data?"""
    ffmpeg_args: str | list[str]
    """Args to pass to ffmpeg"""
    ffmpeg_before_args: str | list[str]
    """Args to pass to ffmpeg before the source"""

    def __init__(self, src: Union[str, Path]) -> None:
        self.source = src
        self.needs_encode = True
        self.locked_stream = False
        self.process: Optional[subprocess.Popen] = None

        self.buffer = AudioBuffer()

        self.buffer_seconds = 3
        self.read_ahead_task = threading.Thread(target=self._read_ahead, daemon=True)

        self.ffmpeg_before_args = ""
        self.ffmpeg_args = ""
        self.probe: bool = False

    def __repr__(self) -> str:
        return f"<{type(self).__name__}: {self.source}>"

    @property
    def _max_buffer_size(self) -> int:
        # 1ms of audio * (buffer seconds * 1000)
        return 192 * (self.buffer_seconds * 1000)

    @property
    def audio_complete(self) -> bool:
        """Uses the state of the subprocess to determine if more audio is coming"""
        return not self.process or self.process.poll() is not None

    def _create_process(self, *, block: bool = True) -> None:
        before = (
            self.ffmpeg_before_args if isinstance(self.ffmpeg_before_args, list) else self.ffmpeg_before_args.split()
        )

        config = {
            "sample_rate": 48000,
            "bitrate": None,
        }

        if shutil.which("ffprobe") is not None and self.probe:
            ffprobe_cmd = [
                "ffprobe",
                "-loglevel",
                "quiet",
                "-print_format",
                "json",
                "-show_streams",
                "-select_streams",
                "a:0",
                self.source,
            ]
            raw_output = subprocess.check_output(ffprobe_cmd, stderr=subprocess.DEVNULL)
            output = FastJson.loads(raw_output)

            config["sample_rate"] = int(output["streams"][0]["sample_rate"])
            config["bitrate"] = int(output["streams"][0]["bit_rate"])

            get_logger().debug(f"Detected audio data for {self.source} - {config}")

            if getattr(self, "bitrate", None) is None and self.encoder:
                self.bitrate = int(config["bitrate"] / 1024)
                self.encoder.set_bitrate(self.bitrate)

        after = self.ffmpeg_args if isinstance(self.ffmpeg_args, list) else self.ffmpeg_args.split()
        cmd = [
            "ffmpeg",
            "-i",
            self.source,
            "-f",
            "s16le",
            "-ar",
            str(config["sample_rate"]),
            "-ac",
            "2",
            "-loglevel",
            "warning",
            "pipe:1",
            "-vn",
        ]
        cmd[1:1] = before
        cmd.extend(after)

        self.process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.DEVNULL)
        self.read_ahead_task.start()

        if block:
            # block until some data is in the buffer
            self.buffer.initialised.wait()

    def _read_ahead(self) -> None:
        while self.process:
            if self.process.poll() is not None:
                # ffmpeg has exited, stop reading ahead
                if not self.buffer.initialised.is_set():
                    # assume this is a small file and initialise the buffer
                    self.buffer.initialised.set()

                return
            if len(self.buffer) < self._max_buffer_size:
                self.buffer.extend(self.process.stdout.read(3840))
            else:
                if not self.buffer.initialised.is_set():
                    self.buffer.initialised.set()
                time.sleep(0.1)

    def pre_buffer(self, duration: None | float = None) -> None:
        """
        Start pre-buffering the audio.

        Args:
            duration: The duration of audio to pre-buffer.
        """
        if duration:
            self.buffer_seconds = duration

        if self.process and self.process.poll() is None:
            raise RuntimeError("Cannot pre-buffer an already running process")
        # sanity value enforcement to prevent audio weirdness
        self.buffer = AudioBuffer()
        self.buffer.initialised.clear()

        self._create_process(block=False)

    def read(self, frame_size: int) -> bytes:
        """
        Reads frame_size bytes of audio from the buffer.

        Returns:
            bytes of audio
        """
        if not self.process:
            self._create_process()
        if not self.buffer.initialised.is_set():
            # we cannot start playing until the buffer is initialised
            self.buffer.initialised.wait()

        data = self.buffer.read(frame_size)

        if len(data) != frame_size:
            data = b""

        return bytes(data)

    def cleanup(self) -> None:
        """Cleans up after this audio object."""
        if self.process and self.process.poll() is None:
            self.process.kill()
            self.process.wait()


class AudioVolume(Audio):
    """An audio object with volume control"""

    _volume: float
    """The internal volume level of the audio"""

    def __init__(self, src: Union[str, Path]) -> None:
        super().__init__(src)
        self._volume = 0.5

    @property
    def volume(self) -> float:
        """The volume of the audio"""
        return self._volume

    @volume.setter
    def volume(self, value: float) -> None:
        """Sets the volume of the audio. Volume cannot be negative."""
        self._volume = max(value, 0.0)

    def read(self, frame_size: int) -> bytes:
        """
        Reads frame_size ms of audio from source.

        Returns:
            bytes of audio
        """
        data = super().read(frame_size)
        return audioop.mul(data, 2, self._volume)



from .context import HybridContext
from .hybrid_slash import HybridSlashCommand, hybrid_slash_command, hybrid_slash_subcommand
from .manager import HybridManager, setup

__all__ = (
    "HybridContext",
    "HybridManager",
    "HybridSlashCommand",
    "hybrid_slash_command",
    "hybrid_slash_subcommand",
    "setup",
)



from abc import ABC, abstractmethod
from datetime import datetime, timedelta, timezone

__all__ = ("BaseTrigger", "IntervalTrigger", "DateTrigger", "TimeTrigger", "OrTrigger")


class BaseTrigger(ABC):
    last_call_time: datetime

    def __new__(cls, *args, **kwargs) -> "BaseTrigger":
        new_cls = super().__new__(cls)
        new_cls.last_call_time = datetime.now()
        return new_cls

    def __or__(self, other: "BaseTrigger") -> "OrTrigger":
        return OrTrigger(self, other)

    def reschedule(self) -> None:
        """Update the last call time to now"""
        self.last_call_time = datetime.now()

    def set_last_call_time(self, call_time: datetime) -> None:
        self.last_call_time = call_time

    @abstractmethod
    def next_fire(self) -> datetime | None:
        """
        Return the next datetime to fire on.

        Returns:
            Datetime if one can be determined. If no datetime can be determined, return None

        """
        ...


class IntervalTrigger(BaseTrigger):
    """
    Trigger the task every set interval.

    Attributes:
        seconds Union[int, float]: How many seconds between intervals
        minutes Union[int, float]: How many minutes between intervals
        hours Union[int, float]: How many hours between intervals
        days Union[int, float]: How many days between intervals
        weeks Union[int, float]: How many weeks between intervals

    """

    _t = int | float

    def __init__(self, seconds: _t = 0, minutes: _t = 0, hours: _t = 0, days: _t = 0, weeks: _t = 0) -> None:
        self.delta = timedelta(days=days, seconds=seconds, minutes=minutes, hours=hours, weeks=weeks)

        # lazy check for negatives
        if (datetime.now() + self.delta) < datetime.now():
            raise ValueError("Interval values must result in a time in the future!")

    def next_fire(self) -> datetime | None:
        return self.last_call_time + self.delta


class DateTrigger(BaseTrigger):
    """
    Trigger the task once, when the specified datetime is reached.

    Attributes:
        target_datetime datetime: A datetime representing the date/time to run this task

    """

    def __init__(self, target_datetime: datetime) -> None:
        self.target = target_datetime

    def next_fire(self) -> datetime | None:
        return self.target if datetime.now() < self.target else None


class TimeTrigger(BaseTrigger):
    """
    Trigger the task every day, at a specified (24 hour clock) time.

    Attributes:
        hour int: The hour of the day (24 hour clock)
        minute int: The minute of the hour
        seconds int: The seconds of the minute
        utc bool: If this time is in UTC

    """

    def __init__(self, hour: int = 0, minute: int = 0, seconds: int = 0, utc: bool = True) -> None:
        self.target_time = (hour, minute, seconds)
        self.tz = timezone.utc if utc else None

    def next_fire(self) -> datetime | None:
        now = datetime.now()
        target = datetime(
            now.year,
            now.month,
            now.day,
            self.target_time[0],
            self.target_time[1],
            self.target_time[2],
            tzinfo=self.tz,
        )
        if target.tzinfo == timezone.utc:
            target = target.astimezone(now.tzinfo)
            target = target.replace(tzinfo=None)

        if target <= self.last_call_time:
            target += timedelta(days=1)
        return target


class OrTrigger(BaseTrigger):
    """Trigger a task when any sub-trigger is fulfilled."""

    def __init__(self, *trigger: BaseTrigger) -> None:
        self.triggers: list[BaseTrigger] = list(trigger)
        self.current_trigger: BaseTrigger = None

    def set_last_call_time(self, call_time: datetime) -> None:
        self.current_trigger.last_call_time = call_time

    def _get_delta(self, d: BaseTrigger) -> timedelta:
        next_fire = d.next_fire()
        return abs(next_fire - self.last_call_time) if next_fire else timedelta.max

    def __or__(self, other: "BaseTrigger") -> "OrTrigger":
        self.triggers.append(other)
        return self

    def _set_current_trigger(self) -> BaseTrigger | None:
        self.current_trigger = self.triggers[0] if len(self.triggers) == 1 else min(self.triggers, key=self._get_delta)
        return self.current_trigger

    def next_fire(self) -> datetime | None:
        return self.current_trigger.next_fire() if self._set_current_trigger() else None



from typing import TYPE_CHECKING, Any, Dict, List, Type

import attrs

from interactions.client.const import T
from interactions.client.mixins.serialization import DictSerializationMixin
from interactions.client.utils.serializer import no_export_meta
from interactions.models.discord.snowflake import SnowflakeObject

if TYPE_CHECKING:
    from interactions.client import Client

__all__ = ("ClientObject", "DiscordObject")


@attrs.define(eq=False, order=False, hash=False, slots=False)
class ClientObject(DictSerializationMixin):
    """Serializable object that requires client reference."""

    _client: "Client" = attrs.field(repr=False, metadata=no_export_meta)

    @property
    def client(self) -> "Client":
        return self._client

    @property
    def bot(self) -> "Client":
        return self._client

    @classmethod
    def _process_dict(cls, data: Dict[str, Any], client: "Client") -> Dict[str, Any]:
        return super()._process_dict(data)

    @classmethod
    def from_dict(cls: Type[T], data: Dict[str, Any], client: "Client") -> T:
        data = cls._process_dict(data, client)
        return cls(client=client, **cls._filter_kwargs(data, cls._get_init_keys()))

    @classmethod
    def from_list(cls: Type[T], datas: List[Dict[str, Any]], client: "Client") -> List[T]:
        return [cls.from_dict(data, client) for data in datas]

    def update_from_dict(self, data) -> T:
        data = self._process_dict(data, self._client)
        for key, value in self._filter_kwargs(data, self._get_keys()).items():
            setattr(self, key, value)

        return self


@attrs.define(eq=False, order=False, hash=False, slots=False)
class DiscordObject(SnowflakeObject, ClientObject):
    pass



from .discord import (
    ActionRow,
    Activity,
    ActivityAssets,
    ActivityFlag,
    ActivityParty,
    ActivitySecrets,
    ActivityTimestamps,
    ActivityType,
    AllowedMentions,
    Application,
    ApplicationCommandPermission,
    ApplicationFlags,
    Asset,
    Attachment,
    AuditLog,
    AuditLogChange,
    AuditLogEntry,
    AuditLogEventType,
    AuditLogHistory,
    AutoArchiveDuration,
    AutoModerationAction,
    AutoModRule,
    BaseChannel,
    BaseComponent,
    BaseGuild,
    BaseMessage,
    BaseSelectMenu,
    BaseUser,
    BrandColors,
    BrandColours,
    Button,
    ButtonStyle,
    ChannelFlags,
    ChannelHistory,
    ChannelMention,
    ChannelSelectMenu,
    ChannelType,
    ClientUser,
    Color,
    COLOR_TYPES,
    Colour,
    CommandType,
    ComponentType,
    CustomEmoji,
    DefaultNotificationLevel,
    DefaultReaction,
    DM,
    DMChannel,
    DMGroup,
    Embed,
    EmbedAttachment,
    EmbedAuthor,
    EmbedField,
    EmbedFooter,
    EmbedProvider,
    ExplicitContentFilterLevel,
    File,
    FlatUIColors,
    FlatUIColours,
    ForumLayoutType,
    get_components_ids,
    Guild,
    GuildBan,
    GuildCategory,
    GuildChannel,
    GuildForum,
    GuildForumPost,
    GuildIntegration,
    GuildMedia,
    GuildNews,
    GuildNewsThread,
    GuildPreview,
    GuildPrivateThread,
    GuildPublicThread,
    GuildStageVoice,
    GuildTemplate,
    GuildText,
    GuildVoice,
    GuildWelcome,
    GuildWelcomeChannel,
    GuildWidget,
    GuildWidgetSettings,
    InputText,
    IntegrationExpireBehaviour,
    Intents,
    InteractionPermissionTypes,
    InteractionType,
    InteractiveComponent,
    InvitableMixin,
    Invite,
    InviteTargetType,
    MaterialColors,
    MaterialColours,
    Member,
    MemberFlags,
    MentionableSelectMenu,
    MentionType,
    Message,
    MessageableMixin,
    MessageActivity,
    MessageActivityType,
    MessageFlags,
    MessageInteraction,
    MessageReference,
    MessageType,
    MFALevel,
    Modal,
    NSFWLevel,
    open_file,
    OverwriteType,
    ParagraphText,
    PartialEmoji,
    PermissionOverwrite,
    Permissions,
    PremiumTier,
    PremiumType,
    process_allowed_mentions,
    process_color,
    process_colour,
    process_components,
    process_default_reaction,
    process_embeds,
    process_emoji,
    process_emoji_req_format,
    process_message_payload,
    process_message_reference,
    process_permission_overwrites,
    process_thread_tag,
    Reaction,
    ReactionUsers,
    Role,
    RoleColors,
    RoleColours,
    RoleSelectMenu,
    ScheduledEvent,
    ScheduledEventPrivacyLevel,
    ScheduledEventStatus,
    ScheduledEventType,
    ShortText,
    Snowflake,
    Snowflake_Type,
    SnowflakeObject,
    spread_to_rows,
    StageInstance,
    StagePrivacyLevel,
    Status,
    Sticker,
    StickerFormatType,
    StickerItem,
    StickerPack,
    StickerTypes,
    StringSelectMenu,
    StringSelectOption,
    SystemChannelFlags,
    Team,
    TeamMember,
    TeamMembershipState,
    TextStyles,
    ThreadableMixin,
    ThreadChannel,
    ThreadList,
    ThreadMember,
    ThreadTag,
    Timestamp,
    TimestampStyles,
    to_optional_snowflake,
    to_snowflake,
    to_snowflake_list,
    TYPE_ALL_CHANNEL,
    TYPE_CHANNEL_MAPPING,
    TYPE_COMPONENT_MAPPING,
    TYPE_DM_CHANNEL,
    TYPE_GUILD_CHANNEL,
    TYPE_MESSAGEABLE_CHANNEL,
    TYPE_THREAD_CHANNEL,
    TYPE_VOICE_CHANNEL,
    UPLOADABLE_TYPE,
    User,
    UserFlags,
    UserSelectMenu,
    VerificationLevel,
    VideoQualityMode,
    VoiceRegion,
    VoiceState,
    Webhook,
    WebhookMixin,
    WebhookTypes,
    WebSocketOPCode,
)
from .internal import (
    ActiveVoiceState,
    application_commands_to_dict,
    auto_defer,
    AutocompleteContext,
    AutoDefer,
    BaseChannelConverter,
    BaseCommand,
    BaseContext,
    BaseInteractionContext,
    BaseTrigger,
    Buckets,
    CallbackObject,
    CallbackType,
    ChannelConverter,
    check,
    component_callback,
    ComponentCommand,
    ComponentContext,
    context_menu,
    user_context_menu,
    message_context_menu,
    ContextMenu,
    ContextMenuContext,
    Converter,
    cooldown,
    Cooldown,
    CooldownSystem,
    SlidingWindowSystem,
    ExponentialBackoffSystem,
    LeakyBucketSystem,
    TokenBucketSystem,
    CustomEmojiConverter,
    DateTrigger,
    dm_only,
    DMChannelConverter,
    DMConverter,
    DMGroupConverter,
    Extension,
    global_autocomplete,
    GlobalAutoComplete,
    Greedy,
    guild_only,
    GuildCategoryConverter,
    GuildChannelConverter,
    GuildConverter,
    GuildNewsConverter,
    GuildNewsThreadConverter,
    GuildPrivateThreadConverter,
    GuildPublicThreadConverter,
    GuildStageVoiceConverter,
    GuildTextConverter,
    GuildVoiceConverter,
    has_any_role,
    has_id,
    has_role,
    IDConverter,
    InteractionCommand,
    InteractionContext,
    IntervalTrigger,
    is_owner,
    listen,
    Listener,
    LocalisedDesc,
    LocalisedName,
    LocalizedDesc,
    LocalizedName,
    max_concurrency,
    MaxConcurrency,
    MemberConverter,
    MessageableChannelConverter,
    MessageConverter,
    modal_callback,
    ModalCommand,
    ModalContext,
    MODEL_TO_CONVERTER,
    NoArgumentConverter,
    OptionType,
    OrTrigger,
    PartialEmojiConverter,
    Resolved,
    RoleConverter,
    slash_attachment_option,
    slash_bool_option,
    slash_channel_option,
    slash_command,
    slash_default_member_permission,
    slash_float_option,
    slash_int_option,
    slash_mentionable_option,
    slash_option,
    slash_role_option,
    slash_str_option,
    slash_user_option,
    SlashCommand,
    SlashCommandChoice,
    SlashCommandOption,
    SlashCommandParameter,
    SlashContext,
    SnowflakeConverter,
    subcommand,
    sync_needed,
    Task,
    ThreadChannelConverter,
    TimeTrigger,
    UserConverter,
    VoiceChannelConverter,
    Wait,
)
from .misc import AsyncIterator, Typing

__all__ = (
    "ActionRow",
    "ActiveVoiceState",
    "Activity",
    "ActivityAssets",
    "ActivityFlag",
    "ActivityParty",
    "ActivitySecrets",
    "ActivityTimestamps",
    "ActivityType",
    "AllowedMentions",
    "Application",
    "application_commands_to_dict",
    "ApplicationCommandPermission",
    "ApplicationFlags",
    "Asset",
    "AsyncIterator",
    "Attachment",
    "AuditLog",
    "AuditLogChange",
    "AuditLogEntry",
    "AuditLogEventType",
    "AuditLogHistory",
    "auto_defer",
    "AutoArchiveDuration",
    "AutocompleteContext",
    "AutoDefer",
    "AutoModerationAction",
    "AutoModRule",
    "BaseChannel",
    "BaseChannelConverter",
    "BaseCommand",
    "BaseComponent",
    "BaseContext",
    "BaseGuild",
    "BaseInteractionContext",
    "BaseMessage",
    "BaseSelectMenu",
    "BaseTrigger",
    "BaseUser",
    "BrandColors",
    "BrandColours",
    "Buckets",
    "Button",
    "ButtonStyle",
    "CallbackObject",
    "CallbackType",
    "ChannelConverter",
    "ChannelFlags",
    "ChannelHistory",
    "ChannelMention",
    "ChannelSelectMenu",
    "ChannelType",
    "check",
    "ClientUser",
    "Color",
    "COLOR_TYPES",
    "Colour",
    "CommandType",
    "component_callback",
    "ComponentCommand",
    "ComponentContext",
    "ComponentType",
    "context_menu",
    "ContextMenu",
    "ContextMenuContext",
    "Converter",
    "cooldown",
    "Cooldown",
    "CooldownSystem",
    "SlidingWindowSystem",
    "ExponentialBackoffSystem",
    "LeakyBucketSystem",
    "TokenBucketSystem",
    "CustomEmoji",
    "CustomEmojiConverter",
    "DateTrigger",
    "DefaultNotificationLevel",
    "DefaultReaction",
    "DM",
    "dm_only",
    "DMChannel",
    "DMChannelConverter",
    "DMConverter",
    "DMGroup",
    "DMGroupConverter",
    "Embed",
    "EmbedAttachment",
    "EmbedAuthor",
    "EmbedField",
    "EmbedFooter",
    "EmbedProvider",
    "ExplicitContentFilterLevel",
    "Extension",
    "File",
    "FlatUIColors",
    "FlatUIColours",
    "ForumLayoutType",
    "get_components_ids",
    "global_autocomplete",
    "GlobalAutoComplete",
    "Greedy",
    "Guild",
    "guild_only",
    "GuildBan",
    "GuildCategory",
    "GuildCategoryConverter",
    "GuildChannel",
    "GuildChannelConverter",
    "GuildConverter",
    "GuildForum",
    "GuildForumPost",
    "GuildIntegration",
    "GuildMedia",
    "GuildNews",
    "GuildNewsConverter",
    "GuildNewsThread",
    "GuildNewsThreadConverter",
    "GuildPreview",
    "GuildPrivateThread",
    "GuildPrivateThreadConverter",
    "GuildPublicThread",
    "GuildPublicThreadConverter",
    "GuildStageVoice",
    "GuildStageVoiceConverter",
    "GuildTemplate",
    "GuildText",
    "GuildTextConverter",
    "GuildVoice",
    "GuildVoiceConverter",
    "GuildWelcome",
    "GuildWelcomeChannel",
    "GuildWidget",
    "GuildWidgetSettings",
    "has_any_role",
    "has_id",
    "has_role",
    "IDConverter",
    "InputText",
    "IntegrationExpireBehaviour",
    "Intents",
    "InteractionCommand",
    "InteractionContext",
    "InteractionPermissionTypes",
    "InteractionType",
    "InteractiveComponent",
    "IntervalTrigger",
    "InvitableMixin",
    "Invite",
    "InviteTargetType",
    "is_owner",
    "listen",
    "Listener",
    "LocalisedDesc",
    "LocalisedName",
    "LocalizedDesc",
    "LocalizedName",
    "MaterialColors",
    "MaterialColours",
    "max_concurrency",
    "MaxConcurrency",
    "Member",
    "MemberConverter",
    "MemberFlags",
    "MentionableSelectMenu",
    "MentionType",
    "Message",
    "message_context_menu",
    "MessageableChannelConverter",
    "MessageableMixin",
    "MessageActivity",
    "MessageActivityType",
    "MessageConverter",
    "MessageFlags",
    "MessageInteraction",
    "MessageReference",
    "MessageType",
    "MFALevel",
    "Modal",
    "modal_callback",
    "ModalCommand",
    "ModalContext",
    "MODEL_TO_CONVERTER",
    "NoArgumentConverter",
    "NSFWLevel",
    "open_file",
    "OptionType",
    "OrTrigger",
    "OverwriteType",
    "ParagraphText",
    "PartialEmoji",
    "PartialEmojiConverter",
    "PermissionOverwrite",
    "Permissions",
    "PremiumTier",
    "PremiumType",
    "process_allowed_mentions",
    "process_color",
    "process_colour",
    "process_components",
    "process_default_reaction",
    "process_embeds",
    "process_emoji",
    "process_emoji_req_format",
    "process_message_payload",
    "process_message_reference",
    "process_permission_overwrites",
    "process_thread_tag",
    "Reaction",
    "ReactionUsers",
    "Resolved",
    "Role",
    "RoleColors",
    "RoleColours",
    "RoleConverter",
    "RoleSelectMenu",
    "ScheduledEvent",
    "ScheduledEventPrivacyLevel",
    "ScheduledEventStatus",
    "ScheduledEventType",
    "ShortText",
    "slash_attachment_option",
    "slash_bool_option",
    "slash_channel_option",
    "slash_command",
    "slash_default_member_permission",
    "slash_float_option",
    "slash_int_option",
    "slash_mentionable_option",
    "slash_option",
    "slash_role_option",
    "slash_str_option",
    "slash_user_option",
    "SlashCommand",
    "SlashCommandChoice",
    "SlashCommandOption",
    "SlashCommandParameter",
    "SlashContext",
    "Snowflake",
    "Snowflake_Type",
    "SnowflakeConverter",
    "SnowflakeObject",
    "spread_to_rows",
    "StageInstance",
    "StagePrivacyLevel",
    "Status",
    "Sticker",
    "StickerFormatType",
    "StickerItem",
    "StickerPack",
    "StickerTypes",
    "StringSelectMenu",
    "StringSelectOption",
    "subcommand",
    "sync_needed",
    "SystemChannelFlags",
    "Task",
    "Team",
    "TeamMember",
    "TeamMembershipState",
    "TextStyles",
    "ThreadableMixin",
    "ThreadChannel",
    "ThreadChannelConverter",
    "ThreadList",
    "ThreadMember",
    "ThreadTag",
    "Timestamp",
    "TimestampStyles",
    "TimeTrigger",
    "to_optional_snowflake",
    "to_snowflake",
    "to_snowflake_list",
    "TYPE_ALL_CHANNEL",
    "TYPE_CHANNEL_MAPPING",
    "TYPE_COMPONENT_MAPPING",
    "TYPE_DM_CHANNEL",
    "TYPE_GUILD_CHANNEL",
    "TYPE_MESSAGEABLE_CHANNEL",
    "TYPE_THREAD_CHANNEL",
    "TYPE_VOICE_CHANNEL",
    "Typing",
    "UPLOADABLE_TYPE",
    "User",
    "user_context_menu",
    "UserConverter",
    "UserFlags",
    "UserSelectMenu",
    "VerificationLevel",
    "VideoQualityMode",
    "VoiceChannelConverter",
    "VoiceRegion",
    "VoiceState",
    "Wait",
    "Webhook",
    "WebhookMixin",
    "WebhookTypes",
    "WebSocketOPCode",
)



"""
These are events dispatched by the client. This is intended as a reference so you know what data to expect for each event.

???+ hint "Example Usage"
    To listen to an event, use the `listen` decorator:

    ```python
    from interactions import listen
    from interactions.api.events import ChannelCreate  # or any other event

    @listen(ChannelCreate)
    async def an_event_handler(event: ChannelCreate):
        print(f"Channel created with name: {event.channel.name}")
    ```

    For more information, including other ways to listen to events, see [the events guide](/interactions.py/Guides/10 Events).

!!! warning
    While all of these events are documented, not all of them are used, currently.

"""
import re
import typing
from typing import Any, Optional, TYPE_CHECKING, Type

import attrs

from interactions.api.events.base import BaseEvent, RawGatewayEvent
from interactions.client.utils.attr_utils import docs

if typing.TYPE_CHECKING:
    from interactions import Extension, BaseCommand, Listener

__all__ = (
    "ButtonPressed",
    "Component",
    "Connect",
    "Disconnect",
    "Error",
    "ShardConnect",
    "ShardDisconnect",
    "Login",
    "Ready",
    "Resume",
    "Select",
    "Startup",
    "WebsocketReady",
    "CommandError",
    "ComponentError",
    "AutocompleteError",
    "ModalError",
    "CommandCompletion",
    "ComponentCompletion",
    "AutocompleteCompletion",
    "ModalCompletion",
    "ExtensionLoad",
    "ExtensionUnload",
    "ExtensionCommandParse",
    "CallbackAdded",
)


if TYPE_CHECKING:
    from interactions.models.internal.context import (
        ComponentContext,
        BaseContext,
        AutocompleteContext,
        ModalContext,
    )

_event_reg = re.compile("(?<!^)(?=[A-Z])")


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class Login(BaseEvent):
    """The bot has just logged in."""


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class Connect(BaseEvent):
    """The bot is now connected to the discord Gateway."""


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class Resume(BaseEvent):
    """The bot has resumed its connection to the discord Gateway."""


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class Disconnect(BaseEvent):
    """The bot has just disconnected."""


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class ShardConnect(Connect):
    """A shard just connected to the discord Gateway."""

    shard_id: int = attrs.field(repr=False, metadata=docs("The ID of the shard"))


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class ShardDisconnect(Disconnect):
    """A shard just disconnected."""

    shard_id: int = attrs.field(repr=False, metadata=docs("The ID of the shard"))


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class Startup(BaseEvent):
    """
    The client is now ready for the first time.

    Use this for tasks you want to do upon login, instead of ready, as
    this will only be called once.

    """


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class Ready(BaseEvent):
    """
    The client is now ready.

    !!! note
        Don't use this event for things that must only happen once, on startup, as this event may be called multiple times.
        Instead, use the `Startup` event

    """


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class WebsocketReady(RawGatewayEvent):
    """The gateway has reported that it is ready."""

    data: dict = attrs.field(repr=False, metadata=docs("The data from the ready event"))


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class Component(BaseEvent):
    """Dispatched when a user uses a Component."""

    ctx: "ComponentContext" = attrs.field(repr=False, metadata=docs("The context of the interaction"))


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class ButtonPressed(Component):
    """Dispatched when a user uses a Button."""


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class Select(Component):
    """Dispatched when a user uses a Select."""


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class CommandCompletion(BaseEvent):
    """Dispatched after the library ran any command callback."""

    ctx: "BaseContext" = attrs.field(repr=False, metadata=docs("The command context"))


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class ComponentCompletion(BaseEvent):
    """Dispatched after the library ran any component callback."""

    ctx: "ComponentContext" = attrs.field(repr=False, metadata=docs("The component context"))


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class AutocompleteCompletion(BaseEvent):
    """Dispatched after the library ran any autocomplete callback."""

    ctx: "AutocompleteContext" = attrs.field(repr=False, metadata=docs("The autocomplete context"))


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class ModalCompletion(BaseEvent):
    """Dispatched after the library ran any modal callback."""

    ctx: "ModalContext" = attrs.field(repr=False, metadata=docs("The modal context"))


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class _Error(BaseEvent):
    error: Exception = attrs.field(repr=False, metadata=docs("The error that was encountered"))
    args: tuple[Any] = attrs.field(repr=False, factory=tuple)
    kwargs: dict[str, Any] = attrs.field(repr=False, factory=dict)


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class Error(_Error):
    """Dispatched when the library encounters an error."""

    source: str = attrs.field(repr=False, metadata=docs("The source of the error"))
    ctx: Optional["BaseContext"] = attrs.field(
        repr=False, default=None, metadata=docs("The Context, if one was active")
    )


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class CommandError(_Error):
    """Dispatched when the library encounters an error in a command."""

    ctx: "BaseContext" = attrs.field(repr=False, metadata=docs("The command context"))


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class ComponentError(_Error):
    """Dispatched when the library encounters an error in a component."""

    ctx: "ComponentContext" = attrs.field(repr=False, metadata=docs("The component context"))


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class AutocompleteError(_Error):
    """Dispatched when the library encounters an error in an autocomplete."""

    ctx: "AutocompleteContext" = attrs.field(repr=False, metadata=docs("The autocomplete context"))


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class ModalError(_Error):
    """Dispatched when the library encounters an error in a modal."""

    ctx: "ModalContext" = attrs.field(repr=False, metadata=docs("The modal context"))


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class ExtensionLoad(BaseEvent):
    """Dispatched when an extension is loaded."""

    extension: "Extension" = attrs.field(repr=False)
    """The extension in question"""

    @property
    def metadata(self) -> "Type[Extension.Metadata] | None":
        """The metadata of the extension, if it has any."""
        return self.extension.Metadata or None


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class ExtensionUnload(ExtensionLoad):
    """Dispatched when an extension is unloaded."""

    ...


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class ExtensionCommandParse(ExtensionLoad):
    """Dispatched when an extension is parsed for commands."""

    callables: list[tuple[str, typing.Callable]] = attrs.field(repr=False, default=None)
    """The callables that were parsed for commands"""


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class CallbackAdded(BaseEvent):
    """Dispatched when a callback is added to the client."""

    callback: "BaseCommand | Listener" = attrs.field(repr=False)
    """The callback that was added"""
    extension: "Extension | None" = attrs.field(repr=False, default=None)
    """The extension that the command was added from, if any"""



from datetime import datetime
from typing import TYPE_CHECKING, cast

import discord_typings

from interactions.client.const import Missing, MISSING
from interactions.client.utils.serializer import dict_filter_none
from interactions.models.discord.timestamp import Timestamp
from interactions.models.internal.protocols import CanRequest
from ..route import Route, PAYLOAD_TYPE

__all__ = ("MemberRequests",)


if TYPE_CHECKING:
    from interactions.models.discord.snowflake import Snowflake_Type


class MemberRequests(CanRequest):
    async def get_member(
        self, guild_id: "Snowflake_Type", user_id: "Snowflake_Type"
    ) -> discord_typings.GuildMemberData:
        """
        Get a member of a guild by ID.

        Args:
            guild_id: The id of the guild
            user_id: The user id to grab

        """
        result = await self.request(
            Route("GET", "/guilds/{guild_id}/members/{user_id}", guild_id=guild_id, user_id=user_id)
        )
        return cast(discord_typings.GuildMemberData, result)

    async def list_members(
        self, guild_id: "Snowflake_Type", limit: int = 1, after: "Snowflake_Type | None" = None
    ) -> list[discord_typings.GuildMemberData]:
        """
        List the members of a guild.

        Args:
            guild_id: The ID of the guild
            limit: How many members to get (max 1000)
            after: Get IDs after this snowflake

        """
        payload: PAYLOAD_TYPE = {
            "limit": limit,
            "after": int(after) if after else None,
        }
        payload = dict_filter_none(payload)

        result = await self.request(Route("GET", "/guilds/{guild_id}/members", guild_id=guild_id), params=payload)
        return cast(list[discord_typings.GuildMemberData], result)

    async def search_guild_members(
        self, guild_id: "Snowflake_Type", query: str, limit: int = 1
    ) -> list[discord_typings.GuildMemberData]:
        """
        Search a guild for members who's username or nickname starts with provided string.

        Args:
            guild_id: The ID of the guild to search
            query: The string to search for
            limit: The number of members to return

        """
        result = await self.request(
            Route("GET", "/guilds/{guild_id}/members/search", guild_id=guild_id),
            params={"query": query, "limit": limit},
        )
        return cast(list[discord_typings.GuildMemberData], result)

    async def modify_guild_member(
        self,
        guild_id: "Snowflake_Type",
        user_id: "Snowflake_Type",
        nickname: str | None | Missing = MISSING,
        roles: list["Snowflake_Type"] | None = None,
        mute: bool | None = None,
        deaf: bool | None = None,
        channel_id: "Snowflake_Type | MISSING" = MISSING,
        communication_disabled_until: str | datetime | Timestamp | None | Missing = MISSING,
        reason: str | None = None,
    ) -> discord_typings.GuildMemberData:
        """
        Modify attributes of a guild member.

        Args:
            guild_id: The ID of the guild
            user_id: The ID of the user we're modifying
            nickname: Value to set users nickname to
            roles: Array of role ids the member is assigned
            mute: Whether the user is muted in voice channels. Will throw a 400 if the user is not in a voice channel
            deaf: Whether the user is deafened in voice channels
            channel_id: id of channel to move user to (if they are connected to voice)
            communication_disabled_until: 	when the user's timeout will expire and the user will be able to communicate in the guild again
            reason: An optional reason for the audit log

        Returns:
            The updated member object

        """
        if isinstance(communication_disabled_until, datetime):
            communication_disabled_until = communication_disabled_until.isoformat()

        payload: PAYLOAD_TYPE = {
            "roles": roles,
            "mute": mute,
            "deaf": deaf,
        }
        payload = dict_filter_none(payload)
        if channel_id is not MISSING:
            payload["channel_id"] = channel_id

        if not isinstance(nickname, Missing):
            payload["nick"] = nickname
        if not isinstance(communication_disabled_until, Missing):
            payload["communication_disabled_until"] = communication_disabled_until

        result = await self.request(
            Route("PATCH", "/guilds/{guild_id}/members/{user_id}", guild_id=guild_id, user_id=user_id),
            payload=payload,
            reason=reason,
        )
        return cast(discord_typings.GuildMemberData, result)

    async def modify_current_member(
        self,
        guild_id: "Snowflake_Type",
        nickname: str | None | Missing = MISSING,
        reason: str | None = None,
    ) -> None:
        """
        Modify attributes of the user

        Args:
            guild_id: The ID of the guild to modify current member in
            nickname: The new nickname to apply
            reason: An optional reason for the audit log

        """
        payload: PAYLOAD_TYPE = {"nick": None if isinstance(nickname, Missing) else nickname}
        await self.request(
            Route("PATCH", "/guilds/{guild_id}/members/@me", guild_id=guild_id),
            payload=payload,
            reason=reason,
        )

    async def add_guild_member_role(
        self,
        guild_id: "Snowflake_Type",
        user_id: "Snowflake_Type",
        role_id: "Snowflake_Type",
        reason: str | None = None,
    ) -> None:
        """
        Adds a role to a guild member.

        Args:
            guild_id: The ID of the guild
            user_id: The ID of the user
            role_id: The ID of the role to add
            reason: The reason for this action

        """
        await self.request(
            Route(
                "PUT",
                "/guilds/{guild_id}/members/{user_id}/roles/{role_id}",
                guild_id=guild_id,
                user_id=user_id,
                role_id=role_id,
            ),
            reason=reason,
        )

    async def remove_guild_member_role(
        self,
        guild_id: "Snowflake_Type",
        user_id: "Snowflake_Type",
        role_id: "Snowflake_Type",
        reason: str | None = None,
    ) -> None:
        """
        Remove a role from a guild member.

        Args:
            guild_id: The ID of the guild
            user_id: The ID of the user
            role_id: The ID of the role to remove
            reason: The reason for this action

        """
        await self.request(
            Route(
                "DELETE",
                "/guilds/{guild_id}/members/{user_id}/roles/{role_id}",
                guild_id=guild_id,
                user_id=user_id,
                role_id=role_id,
            ),
            reason=reason,
        )



import abc
import datetime
import re
import typing
from typing_extensions import Self

import discord_typings
from aiohttp import FormData

from interactions.client import const
from interactions.client.const import get_logger, MISSING
from interactions.models.discord.components import BaseComponent
from interactions.models.discord.file import UPLOADABLE_TYPE
from interactions.models.discord.sticker import Sticker
from interactions.models.discord.user import Member, User

from interactions.models.internal.command import BaseCommand
from interactions.client.mixins.modal import ModalMixin

from interactions.client.errors import HTTPException, AlreadyDeferred, AlreadyResponded
from interactions.client.mixins.send import SendMixin
from interactions.models.discord.enums import (
    Permissions,
    MessageFlags,
    InteractionType,
    ComponentType,
    CommandType,
)
from interactions.models.discord.message import (
    AllowedMentions,
    Attachment,
    Message,
    MessageReference,
    process_message_payload,
)
from interactions.models.discord.snowflake import Snowflake, Snowflake_Type, to_snowflake, to_optional_snowflake
from interactions.models.discord.embed import Embed
from interactions.models.internal.application_commands import (
    OptionType,
    CallbackType,
    SlashCommandChoice,
    SlashCommandOption,
    InteractionCommand,
)

__all__ = (
    "AutocompleteContext",
    "BaseContext",
    "BaseInteractionContext",
    "ComponentContext",
    "ContextMenuContext",
    "InteractionContext",
    "ModalContext",
    "Resolved",
    "SlashContext",
)


if typing.TYPE_CHECKING:
    import interactions


class Resolved:
    """
    A class representing the resolved data from an interaction.

    Attributes:
        channels: A dictionary of channels resolved from the interaction.
        members: A dictionary of members resolved from the interaction.
        users: A dictionary of users resolved from the interaction.
        roles: A dictionary of roles resolved from the interaction.
        messages: A dictionary of messages resolved from the interaction.
        attachments: A dictionary of attachments resolved from the interaction.
    """

    def __init__(self) -> None:
        self.channels: dict[Snowflake, "interactions.TYPE_MESSAGEABLE_CHANNEL"] = {}
        self.members: dict[Snowflake, "interactions.Member"] = {}
        self.users: dict[Snowflake, "interactions.User"] = {}
        self.roles: dict[Snowflake, "interactions.Role"] = {}
        self.messages: dict[Snowflake, "interactions.Message"] = {}
        self.attachments: dict[Snowflake, "interactions.Attachment"] = {}

    def __bool__(self) -> bool:
        """Returns whether any resolved data is present."""
        return (
            bool(self.channels)
            or bool(self.members)
            or bool(self.users)
            or bool(self.roles)
            or bool(self.messages)
            or bool(self.attachments)
        )

    def get(self, snowflake: Snowflake | str, default: typing.Any = None) -> typing.Any:
        snowflake = Snowflake(snowflake)
        """Returns the value of the given snowflake."""
        if channel := self.channels.get(snowflake):
            return channel
        if member := self.members.get(snowflake):
            return member
        if user := self.users.get(snowflake):
            return user
        if role := self.roles.get(snowflake):
            return role
        if message := self.messages.get(snowflake):
            return message
        if attachment := self.attachments.get(snowflake):
            return attachment
        return default

    @classmethod
    def from_dict(cls, client: "interactions.Client", data: dict, guild_id: None | Snowflake = None) -> Self:
        instance = cls()

        if channels := data.get("channels"):
            for key, _channel in channels.items():
                instance.channels[Snowflake(key)] = client.cache.place_channel_data(_channel)

        if members := data.get("members"):
            for key, _member in members.items():
                instance.members[Snowflake(key)] = client.cache.place_member_data(
                    guild_id, {**_member, "user": {**data["users"][key]}}
                )

        if users := data.get("users"):
            for key, _user in users.items():
                instance.users[Snowflake(key)] = client.cache.place_user_data(_user)

        if roles := data.get("roles"):
            for key, _role in roles.items():
                instance.roles[Snowflake(key)] = client.cache.get_role(Snowflake(key))

        if messages := data.get("messages"):
            for key, _msg in messages.items():
                instance.messages[Snowflake(key)] = client.cache.place_message_data(_msg)

        if attachments := data.get("attachments"):
            for key, _attach in attachments.items():
                instance.attachments[Snowflake(key)] = Attachment.from_dict(_attach, client)

        return instance


class BaseContext(metaclass=abc.ABCMeta):
    """
    Base context class for all contexts.

    Define your own context class by inheriting from this class. For compatibility with the library, you must define a `from_dict` classmethod that takes a dict and returns an instance of your context class.

    """

    client: "interactions.Client"
    """The client that created this context."""

    command: BaseCommand
    """The command this context invokes."""

    author_id: Snowflake
    """The id of the user that invoked this context."""
    channel_id: Snowflake
    """The id of the channel this context was invoked in."""
    message_id: Snowflake
    """The id of the message that invoked this context."""

    guild_id: typing.Optional[Snowflake]
    """The id of the guild this context was invoked in, if any."""

    def __init__(self, client: "interactions.Client") -> None:
        self.client = client
        self.author_id = MISSING
        self.channel_id = MISSING
        self.message_id = MISSING
        self.guild_id = MISSING

    @property
    def guild(self) -> typing.Optional["interactions.Guild"]:
        """The guild this context was invoked in."""
        return self.client.cache.get_guild(self.guild_id)

    @property
    def user(self) -> "interactions.User":
        """The user that invoked this context."""
        return self.client.cache.get_user(self.author_id)

    @property
    def member(self) -> typing.Optional["interactions.Member"]:
        """The member object that invoked this context."""
        return self.client.cache.get_member(self.guild_id, self.author_id)

    @property
    def author(self) -> "interactions.Member | interactions.User":
        """The member or user that invoked this context."""
        return self.member or self.user

    @property
    def channel(self) -> "interactions.TYPE_MESSAGEABLE_CHANNEL":
        """The channel this context was invoked in."""
        if self.guild_id:
            return self.client.cache.get_channel(self.channel_id)
        return self.client.cache.get_dm_channel(self.author_id)

    @property
    def message(self) -> typing.Optional["interactions.Message"]:
        """The message that invoked this context, if any."""
        return self.client.cache.get_message(self.channel_id, self.message_id)

    @property
    def voice_state(self) -> typing.Optional["interactions.VoiceState"]:
        """The current voice state of the bot in the guild this context was invoked in, if any."""
        return self.client.cache.get_bot_voice_state(self.guild_id)

    @property
    def bot(self) -> "interactions.Client":
        return self.client

    @classmethod
    @abc.abstractmethod
    def from_dict(cls, client: "interactions.Client", payload: dict) -> Self:
        """
        Create a context instance from a dict.

        Args:
            client: The client creating this context.
            payload: The dict to create the context from.

        Returns:
            The context instance.

        """
        raise NotImplementedError


class BaseInteractionContext(BaseContext):
    token: str
    """The interaction token."""
    id: Snowflake
    """The interaction ID."""

    locale: str
    """The selected locale of the invoking user (https://discord.com/developers/docs/reference#locales)"""
    guild_locale: str
    """The selected locale of the invoking user's guild (https://discord.com/developers/docs/reference#locales)"""
    resolved: Resolved
    """The resolved data for this interaction."""

    # state info
    deferred: bool
    """Whether the interaction has been deferred."""
    responded: bool
    """Whether the interaction has been responded to."""
    ephemeral: bool
    """Whether the interaction response is ephemeral."""

    _context_type: int
    """The context type of the interaction."""
    command_id: Snowflake
    """The command ID of the interaction."""
    _command_name: str
    """The command name of the interaction."""

    permission_map: dict[Snowflake, Permissions]

    args: list[typing.Any]
    """The arguments passed to the interaction."""
    kwargs: dict[str, typing.Any]
    """The keyword arguments passed to the interaction."""

    def __init__(self, client: "interactions.Client") -> None:
        super().__init__(client)
        self.deferred = False
        self.responded = False
        self.ephemeral = False

    @classmethod
    def from_dict(cls, client: "interactions.Client", payload: dict) -> Self:
        instance = cls(client=client)
        instance.token = payload["token"]
        instance.id = Snowflake(payload["id"])
        instance.permission_map = {client.app.id: Permissions(payload.get("app_permissions", 0))}
        instance.locale = payload["locale"]
        instance.guild_locale = payload.get("guild_locale", instance.locale)
        instance._context_type = payload.get("type", 0)
        instance.resolved = Resolved.from_dict(client, payload["data"].get("resolved", {}), payload.get("guild_id"))

        instance.channel_id = Snowflake(payload["channel_id"])
        if member := payload.get("member"):
            instance.author_id = Snowflake(member["user"]["id"])
            instance.guild_id = Snowflake(payload["guild_id"])
            client.cache.place_member_data(instance.guild_id, member)
        else:
            instance.author_id = Snowflake(payload["user"]["id"])
            client.cache.place_user_data(payload["user"])

        if message_data := payload.get("message"):
            message = client.cache.place_message_data(message_data)
            instance.message_id = message.id

        instance.guild_id = to_optional_snowflake(payload.get("guild_id"))

        if payload["type"] in (InteractionType.APPLICATION_COMMAND, InteractionType.AUTOCOMPLETE):
            instance.command_id = Snowflake(payload["data"]["id"])
            instance._command_name = payload["data"]["name"]

        instance.process_options(payload)

        if member := payload.get("member"):
            instance.permission_map[Snowflake(member["id"])] = Permissions(member["permissions"])

        return instance

    @property
    def app_permissions(self) -> Permissions:
        """The permissions available to this interaction"""
        return self.permission_map.get(self.client.app.id, Permissions(0))

    @property
    def author_permissions(self) -> Permissions:
        """The permissions available to the author of this interaction"""
        if self.guild_id:
            return self.permission_map.get(self.author_id, Permissions(0))
        return Permissions(0)

    @property
    def command(self) -> InteractionCommand:
        return self.client._interaction_lookup[self._command_name]

    @property
    def expires_at(self) -> typing.Optional[datetime.datetime]:
        """The time at which the interaction expires."""
        if self.responded:
            return self.id.created_at + datetime.timedelta(minutes=15)
        return self.id.created_at + datetime.timedelta(seconds=3)

    @property
    def expired(self) -> bool:
        """Whether the interaction has expired."""
        return datetime.datetime.utcnow() > self.expires_at

    @property
    def invoke_target(self) -> str:
        """The invoke target of the interaction."""
        return self._command_name

    @property
    def deferred_ephemeral(self) -> bool:
        """Whether the interaction has been deferred ephemerally."""
        return self.deferred and self.ephemeral

    def option_processing_hook(self, option: dict) -> typing.Any:
        """
        Hook for extending options processing.

        This is called for each option, before the library processes it. If this returns a value, the library will not process the option further.

        Args:
            option: The option to process.

        Returns:
            The processed option.

        """
        return option

    def process_options(self, data: discord_typings.InteractionCallbackData) -> None:
        if data["type"] not in (InteractionType.APPLICATION_COMMAND, InteractionType.AUTOCOMPLETE):
            self.args = []
            self.kwargs = {}
            return

        def gather_options(_options: list[dict[str, typing.Any]]) -> dict[str, typing.Any]:
            """Recursively gather options from an option list."""
            kwargs = {}
            for option in _options:
                if hook_result := self.option_processing_hook(option):
                    kwargs[option["name"]] = hook_result

                if option["type"] in (OptionType.SUB_COMMAND, OptionType.SUB_COMMAND_GROUP):
                    self._command_name = f"{self._command_name} {option['name']}"
                    return gather_options(option["options"])

                value = option.get("value")

                if option["type"] in OptionType.resolvable_types():
                    value = self.resolved.get(value, value)

                kwargs[option["name"]] = value
            return kwargs

        if options := data["data"].get("options"):
            self.kwargs = gather_options(options)  # type: ignore
        else:
            self.kwargs = {}
        self.args = list(self.kwargs.values())


class InteractionContext(BaseInteractionContext, SendMixin):
    async def defer(self, *, ephemeral: bool = False) -> None:
        """
        Defer the interaction.

        Args:
            ephemeral: Whether the interaction response should be ephemeral.
        """
        if self.deferred:
            raise AlreadyDeferred("Interaction has already been responded to.")
        if self.responded:
            raise AlreadyResponded("Interaction has already been responded to.")

        payload = {"type": CallbackType.DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE}
        if ephemeral:
            payload["data"] = {"flags": MessageFlags.EPHEMERAL}

        await self.client.http.post_initial_response(payload, self.id, self.token)
        self.deferred = True
        self.ephemeral = ephemeral

    async def _send_http_request(
        self, message_payload: dict, files: typing.Iterable["UPLOADABLE_TYPE"] | None = None
    ) -> dict:
        if const.has_client_feature("FOLLOWUP_INTERACTIONS_FOR_IMAGES") and not self.deferred and not self.responded:
            # experimental bypass for discords broken image proxy
            if embeds := message_payload.get("embeds", {}):
                if any(e.get("image") for e in embeds) or any(e.get("thumbnail") for e in embeds):
                    if MessageFlags.EPHEMERAL in message_payload.get("flags", MessageFlags.NONE):
                        self.ephemeral = True
                    await self.defer(ephemeral=self.ephemeral)

        if self.responded:
            message_data = await self.client.http.post_followup(
                message_payload, self.client.app.id, self.token, files=files
            )
        else:
            if isinstance(message_payload, FormData) and not self.deferred:
                await self.defer(ephemeral=self.ephemeral)
            if self.deferred:
                if const.has_client_feature("FOLLOWUP_INTERACTIONS_FOR_IMAGES"):
                    message_data = await self.client.http.post_followup(
                        message_payload, self.client.app.id, self.token, files=files
                    )
                else:
                    message_data = await self.client.http.edit_interaction_message(
                        message_payload, self.client.app.id, self.token, files=files
                    )
            else:
                payload = {
                    "type": CallbackType.CHANNEL_MESSAGE_WITH_SOURCE,
                    "data": message_payload,
                }
                message_data = await self.client.http.post_initial_response(payload, self.id, self.token, files=files)

        if not message_data:
            try:
                message_data = await self.client.http.get_interaction_message(self.client.app.id, self.token)
            except HTTPException:
                pass

        self.responded = True
        return message_data

    async def send(
        self,
        content: typing.Optional[str] = None,
        *,
        embeds: typing.Optional[
            typing.Union[typing.Iterable[typing.Union["Embed", dict]], typing.Union["Embed", dict]]
        ] = None,
        embed: typing.Optional[typing.Union["Embed", dict]] = None,
        components: typing.Optional[
            typing.Union[
                typing.Iterable[typing.Iterable[typing.Union["BaseComponent", dict]]],
                typing.Iterable[typing.Union["BaseComponent", dict]],
                "BaseComponent",
                dict,
            ]
        ] = None,
        stickers: typing.Optional[
            typing.Union[
                typing.Iterable[typing.Union["Sticker", "Snowflake_Type"]],
                "Sticker",
                "Snowflake_Type",
            ]
        ] = None,
        allowed_mentions: typing.Optional[typing.Union["AllowedMentions", dict]] = None,
        reply_to: typing.Optional[typing.Union["MessageReference", "Message", dict, "Snowflake_Type"]] = None,
        files: typing.Optional[typing.Union["UPLOADABLE_TYPE", typing.Iterable["UPLOADABLE_TYPE"]]] = None,
        file: typing.Optional["UPLOADABLE_TYPE"] = None,
        tts: bool = False,
        suppress_embeds: bool = False,
        silent: bool = False,
        flags: typing.Optional[typing.Union[int, "MessageFlags"]] = None,
        delete_after: typing.Optional[float] = None,
        ephemeral: bool = False,
        **kwargs: typing.Any,
    ) -> "interactions.Message":
        """
        Send a message.

        Args:
            content: Message text content.
            embeds: Embedded rich content (up to 6000 characters).
            embed: Embedded rich content (up to 6000 characters).
            components: The components to include with the message.
            stickers: IDs of up to 3 stickers in the server to send in the message.
            allowed_mentions: Allowed mentions for the message.
            reply_to: Message to reference, must be from the same channel.
            files: Files to send, the path, bytes or File() instance, defaults to None. You may have up to 10 files.
            file: Files to send, the path, bytes or File() instance, defaults to None. You may have up to 10 files.
            tts: Should this message use Text To Speech.
            suppress_embeds: Should embeds be suppressed on this send
            silent: Should this message be sent without triggering a notification.
            flags: Message flags to apply.
            delete_after: Delete message after this many seconds.
            ephemeral: Whether the response should be ephemeral

        Returns:
            New message object that was sent.
        """
        flags = MessageFlags(flags or 0)
        if ephemeral:
            flags |= MessageFlags.EPHEMERAL
            self.ephemeral = True
        if suppress_embeds:
            flags |= MessageFlags.SUPPRESS_EMBEDS
        if silent:
            flags |= MessageFlags.SILENT

        return await super().send(
            content=content,
            embeds=embeds,
            embed=embed,
            components=components,
            stickers=stickers,
            allowed_mentions=allowed_mentions,
            reply_to=reply_to,
            files=files,
            file=file,
            tts=tts,
            flags=flags,
            delete_after=delete_after,
            pass_self_into_delete=True,
            **kwargs,
        )

    respond = send

    async def delete(self, message: "Snowflake_Type") -> None:
        """
        Delete a message sent in response to this interaction.

        Args:
            message: The message to delete
        """
        await self.client.http.delete_interaction_message(self.client.app.id, self.token, to_snowflake(message))

    async def edit(
        self,
        message: "Snowflake_Type",
        *,
        content: typing.Optional[str] = None,
        embeds: typing.Optional[
            typing.Union[typing.Iterable[typing.Union["Embed", dict]], typing.Union["Embed", dict]]
        ] = None,
        embed: typing.Optional[typing.Union["Embed", dict]] = None,
        components: typing.Optional[
            typing.Union[
                typing.Iterable[typing.Iterable[typing.Union["BaseComponent", dict]]],
                typing.Iterable[typing.Union["BaseComponent", dict]],
                "BaseComponent",
                dict,
            ]
        ] = None,
        attachments: typing.Optional[typing.Sequence[Attachment | dict]] = None,
        allowed_mentions: typing.Optional[typing.Union["AllowedMentions", dict]] = None,
        files: typing.Optional[typing.Union["UPLOADABLE_TYPE", typing.Iterable["UPLOADABLE_TYPE"]]] = None,
        file: typing.Optional["UPLOADABLE_TYPE"] = None,
        tts: bool = False,
    ) -> "interactions.Message":
        message_payload = process_message_payload(
            content=content,
            embeds=embeds or embed,
            components=components,
            allowed_mentions=allowed_mentions,
            attachments=attachments,
            tts=tts,
        )

        if file:
            files = [file, *files] if files else [file]
        message_data = await self.client.http.edit_interaction_message(
            payload=message_payload,
            application_id=self.client.app.id,
            token=self.token,
            message_id=to_snowflake(message),
            files=files,
        )
        if message_data:
            return self.client.cache.place_message_data(message_data)


class SlashContext(InteractionContext, ModalMixin):
    @classmethod
    def from_dict(cls, client: "interactions.Client", payload: dict) -> Self:
        return super().from_dict(client, payload)


class ContextMenuContext(InteractionContext, ModalMixin):
    target_id: Snowflake
    """The id of the target of the context menu."""
    editing_origin: bool
    """Whether you have deferred the interaction and are editing the original response."""
    target_type: None | CommandType
    """The type of the target of the context menu."""

    def __init__(self, client: "interactions.Client") -> None:
        super().__init__(client)
        self.editing_origin = False

    @classmethod
    def from_dict(cls, client: "interactions.Client", payload: dict) -> Self:
        instance = super().from_dict(client, payload)
        instance.target_id = Snowflake(payload["data"]["target_id"])
        instance.target_type = CommandType(payload["data"]["type"])
        return instance

    async def defer(self, *, ephemeral: bool = False, edit_origin: bool = False) -> None:
        """
        Defer the interaction.

        Args:
            ephemeral: Whether the interaction response should be ephemeral.
            edit_origin: Whether to edit the original message instead of sending a new one.
        """
        if self.deferred:
            raise AlreadyDeferred("Interaction has already been responded to.")
        if self.responded:
            raise AlreadyResponded("Interaction has already been responded to.")

        payload = {
            "type": CallbackType.DEFERRED_UPDATE_MESSAGE
            if edit_origin
            else CallbackType.DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE
        }
        if ephemeral:
            if edit_origin:
                raise ValueError("Cannot use ephemeral and edit_origin together.")
            payload["data"] = {"flags": MessageFlags.EPHEMERAL}

        await self.client.http.post_initial_response(payload, self.id, self.token)
        self.deferred = True
        self.ephemeral = ephemeral
        self.editing_origin = edit_origin

    @property
    def target(self) -> None | Message | User | Member:
        """
        The target of the context menu.

        Returns:
            The target of the context menu.
        """
        return self.resolved.get(self.target_id)


class ComponentContext(InteractionContext, ModalMixin):
    values: list[str]
    """The values of the SelectMenu component, if any."""
    custom_id: str
    """The custom_id of the component."""
    component_type: int
    """The type of the component."""
    defer_edit_origin: bool
    """Whether you have deferred the interaction and are editing the original response."""

    @classmethod
    def from_dict(cls, client: "interactions.Client", payload: dict) -> Self:
        instance = super().from_dict(client, payload)
        instance.values = payload["data"].get("values", [])
        instance.custom_id = payload["data"]["custom_id"]
        instance._command_id = instance.custom_id
        instance._command_name = instance.custom_id
        instance.component_type = payload["data"]["component_type"]
        instance.defer_edit_origin = False

        searches = {
            "users": instance.component_type in (ComponentType.USER_SELECT, ComponentType.MENTIONABLE_SELECT),
            "members": instance.guild_id
            and instance.component_type in (ComponentType.USER_SELECT, ComponentType.MENTIONABLE_SELECT),
            "channels": instance.component_type in (ComponentType.CHANNEL_SELECT, ComponentType.MENTIONABLE_SELECT),
            "roles": instance.guild_id
            and instance.component_type in (ComponentType.ROLE_SELECT, ComponentType.MENTIONABLE_SELECT),
        }

        if instance.component_type in (
            ComponentType.USER_SELECT,
            ComponentType.CHANNEL_SELECT,
            ComponentType.ROLE_SELECT,
            ComponentType.MENTIONABLE_SELECT,
        ):
            for i, value in enumerate(instance.values):
                if re.match(r"\d{17,}", value):
                    key = Snowflake(value)

                    if resolved := instance.resolved.get(key):
                        instance.values[i] = resolved
                    elif searches["members"] and (member := instance.client.cache.get_member(instance.guild_id, key)):
                        instance.values[i] = member
                    elif searches["users"] and (user := instance.client.cache.get_user(key)):
                        instance.values[i] = user
                    elif searches["roles"] and (role := instance.client.cache.get_role(key)):
                        instance.values[i] = role
                    elif searches["channels"] and (channel := instance.client.cache.get_channel(key)):
                        instance.values[i] = channel
        return instance

    async def defer(self, *, ephemeral: bool = False, edit_origin: bool = False) -> None:
        """
        Defer the interaction.

        Args:
            ephemeral: Whether the interaction response should be ephemeral.
            edit_origin: Whether to edit the original message instead of sending a new one.
        """
        if self.deferred:
            raise AlreadyDeferred("Interaction has already been responded to.")
        if self.responded:
            raise AlreadyResponded("Interaction has already been responded to.")

        payload = {
            "type": CallbackType.DEFERRED_UPDATE_MESSAGE
            if edit_origin
            else CallbackType.DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE
        }
        if ephemeral:
            if edit_origin:
                raise ValueError("Cannot use ephemeral and edit_origin together.")
            payload["data"] = {"flags": MessageFlags.EPHEMERAL}

        await self.client.http.post_initial_response(payload, self.id, self.token)
        self.deferred = True
        self.ephemeral = ephemeral
        self.editing_origin = edit_origin

    async def edit_origin(
        self,
        *,
        content: typing.Optional[str] = None,
        embeds: typing.Optional[
            typing.Union[typing.Iterable[typing.Union["Embed", dict]], typing.Union["Embed", dict]]
        ] = None,
        embed: typing.Optional[typing.Union["Embed", dict]] = None,
        components: typing.Optional[
            typing.Union[
                typing.Iterable[typing.Iterable[typing.Union["BaseComponent", dict]]],
                typing.Iterable[typing.Union["BaseComponent", dict]],
                "BaseComponent",
                dict,
            ]
        ] = None,
        allowed_mentions: typing.Optional[typing.Union["AllowedMentions", dict]] = None,
        files: typing.Optional[typing.Union["UPLOADABLE_TYPE", typing.Iterable["UPLOADABLE_TYPE"]]] = None,
        file: typing.Optional["UPLOADABLE_TYPE"] = None,
        tts: bool = False,
    ) -> "Message":
        """
        Edits the original message of the component.

        Args:
            content: Message text content.
            embeds: Embedded rich content (up to 6000 characters).
            embed: Embedded rich content (up to 6000 characters).
            components: The components to include with the message.
            allowed_mentions: Allowed mentions for the message.
            files: Files to send, the path, bytes or File() instance, defaults to None. You may have up to 10 files.
            file: Files to send, the path, bytes or File() instance, defaults to None. You may have up to 10 files.
            tts: Should this message use Text To Speech.

        Returns:
            The message after it was edited.
        """
        if not self.responded and not self.deferred and (files or file):
            # Discord doesn't allow files at initial response, so we defer then edit.
            await self.defer(edit_origin=True)

        message_payload = process_message_payload(
            content=content,
            embeds=embeds or embed,
            components=components,
            allowed_mentions=allowed_mentions,
            tts=tts,
        )

        message_data = None
        if self.deferred:
            if not self.editing_origin:
                get_logger().warning(
                    "If you want to edit the original message, and need to defer, you must set the `edit_origin` kwarg to True!"
                )

            message_data = await self.client.http.edit_interaction_message(
                message_payload, self.client.app.id, self.token, files=files or file
            )
            self.deferred = False
            self.editing_origin = False
        else:
            payload = {"type": CallbackType.UPDATE_MESSAGE, "data": message_payload}
            await self.client.http.post_initial_response(payload, str(self.id), self.token, files=files or file)
            message_data = await self.client.http.get_interaction_message(self.client.app.id, self.token)

        if message_data:
            message = self.client.cache.place_message_data(message_data)
            self.message_id = message.id
            return message

    @property
    def component(self) -> typing.Optional[BaseComponent]:
        """The component that was interacted with."""
        if self.message is None or self.message.components is None:
            return None
        for action_row in self.message.components:
            for component in action_row.components:
                if component.custom_id == self.custom_id:
                    return component


class ModalContext(InteractionContext):
    responses: dict[str, str]
    """The responses of the modal. The key is the `custom_id` of the component."""
    custom_id: str
    """The developer defined custom ID of this modal"""
    edit_origin: bool
    """Whether to edit the original message instead of sending a new one."""

    @classmethod
    def from_dict(cls, client: "interactions.Client", payload: dict) -> Self:
        instance = super().from_dict(client, payload)
        instance.responses = {
            comp["components"][0]["custom_id"]: comp["components"][0]["value"] for comp in payload["data"]["components"]
        }
        instance.kwargs = instance.responses
        instance.custom_id = payload["data"]["custom_id"]
        instance.edit_origin = False
        return instance

    async def edit(self, message: "Snowflake_Type", **kwargs) -> "interactions.Message":
        if not self.deferred and not self.responded:
            await self.defer(edit_origin=True)
        return await super().edit(message, **kwargs)

    async def defer(self, *, ephemeral: bool = False, edit_origin: bool = False) -> None:
        """
        Defer the interaction.

        Args:
            ephemeral: Whether the interaction response should be ephemeral.
            edit_origin: Whether to edit the original message instead of sending a followup.
        """
        if self.deferred:
            raise AlreadyDeferred("Interaction has already been responded to.")
        if self.responded:
            raise AlreadyResponded("Interaction has already been responded to.")

        payload = {
            "type": CallbackType.DEFERRED_UPDATE_MESSAGE
            if edit_origin
            else CallbackType.DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE
        }
        if ephemeral:
            payload["data"] = {"flags": MessageFlags.EPHEMERAL}

        if edit_origin:
            self.edit_origin = True

        await self.client.http.post_initial_response(payload, self.id, self.token)
        self.deferred = True
        self.ephemeral = ephemeral


class AutocompleteContext(BaseInteractionContext):
    focussed_option: SlashCommandOption  # todo: option parsing
    """The option the user is currently filling in."""

    @classmethod
    def from_dict(cls, client: "interactions.Client", payload: dict) -> Self:
        return super().from_dict(client, payload)

    @property
    def input_text(self) -> str:
        """The text the user has already filled in."""
        return self.kwargs.get(str(self.focussed_option.name), "")

    def option_processing_hook(self, option: dict) -> None:
        if option.get("focused", False):
            self.focussed_option = SlashCommandOption.from_dict(option)
        return

    async def send(
        self, choices: typing.Iterable[str | int | float | dict[str, int | float | str] | SlashCommandChoice]
    ) -> None:
        """
        Send your autocomplete choices to discord. Choices must be either a list of strings, or a dictionary following the following format:

        ```json
            {
              "name": str,
              "value": str
            }
        ```
        Where name is the text visible in Discord, and value is the data sent back to your client when that choice is
        chosen.

        Args:
            choices: 25 choices the user can pick
        """
        if self.focussed_option.type == OptionType.STRING:
            type_cast = str
        elif self.focussed_option.type == OptionType.INTEGER:
            type_cast = int
        elif self.focussed_option.type == OptionType.NUMBER:
            type_cast = float
        else:
            type_cast = None

        processed_choices = []
        for choice in choices:
            if isinstance(choice, dict):
                name = choice["name"]
                value = choice["value"]
            elif isinstance(choice, SlashCommandChoice):
                name = choice.name.get_locale(self.locale)
                value = choice.value
            else:
                name = str(choice)
                value = choice

            processed_choices.append({"name": name, "value": type_cast(value) if type_cast else value})

        payload = {"type": CallbackType.AUTOCOMPLETE_RESULT, "data": {"choices": processed_choices}}
        await self.client.http.post_initial_response(payload, self.id, self.token)
